#! /usr/local/bin/python3

from argparse import ArgumentParser
from rad_tools.routines import GREEN, RESET, RED, YELLOW
from rad_tools.jarvis.head import Action


def manager(no_log=False):
    do_log = not no_log
    if do_log:
        pass
    print(f"{GREEN}JARVIS is here\n{RESET}")
    action = Action("Initiate")
    print(f"{GREEN}JARVIS is out of here\n{RESET}")


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("-nl", "--no-log",
                        default=False,
                        action="store_true",
                        help="Use if you dont want to keep the log.")
    args = parser.parse_args()

    manager(no_log=args.no_log)



from copy import deepcopy

import numpy as np

from rad_tools.routines import exchange_from_matrix, exchange_to_matrix


class Bond:
    """
    Class with implemented logic for one bond.

    Parameters
    ----------

    iso : int or float, default 0
        Value of isotropic exchange parameter in meV. 

    aniso : 3 x 3 np.ndarray, default np.zeros((3, 3))
        3 x 3 matrix of symmetric anisotropic exchange in meV. 

    dmi : 3 x 1 np.ndarray, default np.zeros(3)
        Dzyaroshinsky-Moria interaction vector (Dx, Dy, Dz) in meV. 

    matrix : 3 x 3 np.ndarray, default np.zeros((3, 3))
        Exchange matrix in meV. If ``matrix`` is specified then ``iso`` ,
        ``aniso`` and ``dmi`` will be ignored and derived from ``matrix`` .
        If ``matrix`` is not specified then it will be derived from
        ``iso`` , ``aniso`` and ``dmi`` .

    distance : float, default 0
        Lenght of the bond.

    Attributes
    ----------

    iso : float
        Value of isotropic exchange parameter in meV. If ``iso`` is
        not specified then it will be 0.

        Matrix form: ::

            [[J, 0, 0],
             [0, J, 0],
             [0, 0, J]]

    aniso : 3 x 3 np.ndarray of floats
        3 x 3 matrix of symmetric anisotropic exchange in meV. If ``aniso``
        is not specified then it will be filled with zeros.

        Matrix form: ::

            [[J_xx, J_xy, J_xz],
             [J_xy, J_yy, J_yz],
             [J_xz, J_yz, J_zz]]

    dmi : 3 x 1 np.ndarray of floats
        Dzyaroshinsky-Moria interaction vector (Dx, Dy, Dz) in meV. If ``dmi``
        is not specified then it will be filled with zeros.

        Vector form: ::

            [D_x, D_y, D_z]

        Matrix form: ::

            [[0, D_z, -D_y],
             [-D_z, 0, D_x],
             [D_y, -D_x, 0]]

    matrix : 3 x 3 np.ndarray of floats
        Exchange matrix in meV. If ``matrix`` is specified then ``iso`` ,
        ``aniso`` and ``dmi`` will be ignored and derived from ``matrix`` .
        If ``matrix`` is not specified then it will be derived from
        ``iso`` , ``aniso`` and ``dmi`` .

        Matrix form ::

            [[J_xx, J_xy, J_xz],
             [J_yx, J_yy, J_yz],
             [J_zx, J_zy, J_zz]]

    distance : float
        Lenght of the bond.

    """

    distance_tolerance = 1E-05

    def __init__(self,
                 iso=0,
                 aniso=np.zeros((3, 3)),
                 dmi=np.zeros(3),
                 matrix=np.zeros((3, 3)),
                 distance=0.) -> None:
        self.iso = 0.
        self.aniso = np.zeros((3, 3), dtype=float)
        self.dmi = np.zeros(3, dtype=float)
        self.matrix = np.zeros((3, 3), dtype=float)
        self.dis = float(distance)

        self.matrix = np.array(matrix, dtype=float)
        if (self.matrix == np.zeros((3, 3))).all():
            self.iso = float(iso)
            self.aniso = np.array(aniso, dtype=float)
            self.dmi = np.array(dmi, dtype=float)
            self.matrix = exchange_to_matrix(self.iso, self.aniso, self.dmi)
        # To ensure the correct decomposition
        self.iso, self.aniso, self.dmi = exchange_from_matrix(self.matrix)

    def __add__(self, other):
        iso = (self.iso + other.iso)
        aniso = (self.aniso + other.aniso)
        dmi = (self.dmi + other.dmi)
        if self.dis - other.dis < self.distance_tolerance:
            dis = self.dis
        else:
            raise ValueError('Two bonds have different distance, could hot add. '
                             f'(Tolerance: {self.distance_tolerance})')
        return Bond(iso=iso, aniso=aniso, dmi=dmi, distance=dis)

    def __sub__(self, other):
        iso = (self.iso - other.iso)
        aniso = (self.aniso - other.aniso)
        dmi = (self.dmi - other.dmi)
        if self.dis - other.dis < self.distance_tolerance:
            dis = self.dis
        else:
            raise ValueError('Two bonds have different distance, '
                             'could hot subtraction. '
                             f'(Tolerance: {self.distance_tolerance})')
        return Bond(iso=iso, aniso=aniso, dmi=dmi, distance=dis)

    def __mul__(self, number):
        iso = self.iso * number
        aniso = self.aniso * number
        dmi = self.dmi * number
        return Bond(iso=iso, aniso=aniso, dmi=dmi, distance=self.dis)

    def __rmul__(self, number):
        iso = number * self.iso
        aniso = number * self.aniso
        dmi = number * self.dmi
        return Bond(iso=iso, aniso=aniso, dmi=dmi, distance=self.dis)


class ExchangeModel:
    """
    Class with the logic for the exchange model.

    Attributes
    ----------

    cell : 3 x 3 np.ndarray of floats
        3 x 3 matrix of lattice vectors in Angstrom.::

            [[a_x  a_y  a_z],
             [b_x  b_y  b_z],
             [c_x  x_y  c_z]]

    magnetic_atoms : dict
       Dictionary with keys : str - marks of atoms and value : 1 x 3 np.ndarray
       - coordinate of the atom in Angstroms.

    bonds : dict
        Dictionary of bonds. ::

            {Atom_1: {Atom_2: {R: bond, ...}, ...}, ...}
    """

    def __init__(self) -> None:
        self.cell = np.zeros((3, 3), dtype=float)
        self.magnetic_atoms = {}
        self.bonds = {}

    def add_atom(self, name, x, y, z):
        """
        Add magnetic atom to the model.

        Parameters
        ----------
        name : str
            Mark for the atom. Note: if an atom with the same mark already
            exists in ``magnetic_atoms`` then it will be rewritten.

        x : int or float
            x coordinate of the atom, in Angstroms.

        y : int or float
            y coordinate of the atom, in Angstroms.

        z : int or float
            z coordinate of the atom, in Angstroms.
        """
        self.magnetic_atoms[name] = (float(x), float(y), float(z))

    def remove_atom(self, name):
        """
        Remove magnetic atom from the model

        Note: this method will remove atom from ``magnetic_atoms`` and all the
        bonds, which starts or ends in this atom.

        Parameters
        ----------
        name : str
            Mark for the atom.
        """
        if name in self.magnetic_atoms:
            del self.magnetic_atoms[name]
        if name in self.bonds:
            del self.bonds[name]
        for atom1 in self.bonds:
            if name in self.bonds[atom1]:
                del self.bonds[atom1][name]

    def add_bond(self, bond, atom1, atom2, R):
        """
        Add one bond to the model.

        Parameters
        ----------
        bond : :py:class:`Bond`
            An instance of :py:class:`Bond` class with the information about
            exchange parameters.
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.
        atom2 : str
            Name of atom2 in R unit cell.
        R : tuple of ints
            Radius vector of the unit cell for atom2.
        """
        if atom1 not in self.bonds:
            self.bonds[atom1] = {}
        if atom2 not in self.bonds[atom1]:
            self.bonds[atom1][atom2] = {}
        self.bonds[atom1][atom2][R] = bond

    def remove_bond(self, atom1, atom2, R):
        """
        Remove one bond from the model.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.

        atom2 : str
            Name of atom2 in R unit cell.

        R : tuple of ints
            Radius vector of the unit cell for atom2.
        """
        if atom1 in self.bonds and \
            atom2 in self.bonds[atom1] and \
                R in self.bonds[atom1][atom2]:
            del self.bonds[atom1][atom2][R]
            if not self.bonds[atom1][atom2]:
                del self.bonds[atom1][atom2]
            if not self.bonds[atom1]:
                del self.bonds[atom1]

    def get_lattice_vectors_length(self):
        """
        Getter for the lattice vectors length.

        Returns
        -------
        a : float
            Length of lattice vector a.

        b : float
            Length of lattice vector b.

        c : float
            Length of lattice vector c.
        """
        a = np.sqrt(np.sum(self.cell[0]**2))
        b = np.sqrt(np.sum(self.cell[1]**2))
        c = np.sqrt(np.sum(self.cell[2]**2))
        return a, b, c

    def get_atom_coordinates(self, atom1, atom2, R):
        """
        Getter for the pair of atom coordinates.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.

        atom2 : str
            Name of atom2 in R unit cell.

        R : tuple of ints
            Radius vector of the unit cell for atom2.

        Returns
        -------
        x1 : float
            x coordinate of atom1 in the cell (0 0 0).

        y1 : float
            y coordinate of atom1 in the cell (0 0 0).

        z1 : float
            z coordinate of atom1 in the cell (0 0 0).

        x2 : float
            x coordinate of atom2 in the cell R.

        y2 : float
            y coordinate of atom2 in the cell R.

        z2 : float
            z coordinate of atom2 in the cell R.
        """
        x1 = self.magnetic_atoms[atom1][0]
        y1 = self.magnetic_atoms[atom1][1]
        z1 = self.magnetic_atoms[atom1][2]

        x2 = np.sum((R * self.cell.T).T, axis=0)[0] +\
            self.magnetic_atoms[atom2][0]
        y2 = np.sum((R * self.cell.T).T, axis=0)[1] +\
            self.magnetic_atoms[atom2][1]
        z2 = np.sum((R * self.cell.T).T, axis=0)[2] +\
            self.magnetic_atoms[atom2][2]
        return x1, y1, z1, x2, y2, z2

    def get_bond_coordinate(self, atom1, atom2, R):
        """
        Getter for the middle point of the bond.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.

        atom2 : str
            Name of atom2 in R unit cell.

        R : tuple of ints
            Radius vector of the unit cell for atom2.

        Returns
        -------
        x : float
            x coordinate of the point in the middle of bond.

        y : float
            y coordinate of the point in the middle of bond.

        z : float
            z coordinate of the point in the middle of bond.

        """
        x1, y1, z1, x2, y2, z2 = self.get_atom_coordinates(atom1, atom2, R)

        return (x1 + x2) / 2, (y1 + y2) / 2, (z1 + z2) / 2

    def get_space_dimensions(self):
        """
        Getter for the sample size.

        Return the maximum coordinate for all 3 axes.
        Maximumm is taken for all directions.

        Returns
        -------
        X : float
            Maximum x coordinate.

        Y : float
            Maximum y coordinate.

        Z : float
            Maximum z coordinate.
        """
        X, Y, Z = 0, 0, 0
        for atom1 in self.bonds:
            for atom2 in self.bonds[atom1]:
                for R in self.bonds[atom1][atom2]:
                    x1, y1, z1, x2, y2, z2 = self.get_atom_coordinates(atom1,
                                                                       atom2,
                                                                       R)
                    X = max(abs(x1), abs(x2), X)
                    Y = max(abs(x1), abs(x2), Y)
                    Z = max(abs(x1), abs(x2), Z)
        return X, Y, Z

    def get_cells(self):
        """
        Getter for the list of cells.

        Return the list of R, specifying all cell that are present in the model.

        Returns
        -------
        cells : list of tuples of ints
            List of R.
        """
        cells = set()
        for atom1 in self.bonds:
            for atom2 in self.bonds[atom1]:
                for R in self.bonds[atom1][atom2]:
                    cells.add(R)
        return list(cells)

    def get_bond_list(self):
        """
        Getter for the list of bonds from the model.

        Returns
        -------
        bond_list : list
            List with all the bonds from the model.::

                [(atom1, atom2, R), ...]
        """
        bond_list = []
        for atom1 in self.bonds:
            for atom2 in self.bonds[atom1]:
                for R in self.bonds[atom1][atom2]:
                    bond_list.append((atom1, atom2, R))
        return bond_list

    # TODO It is ugly, redo in a beautifull way.
    def remove_double_bonds(self):
        """
        Remove double bonds.

        If for atom pair atom1, atom2 exist bond 1-2 and 2-1 they will be merged.
        All values will be calcuated as (1-2 + 2-1) / 2.
        Note: this method is not modifying the instance at which it is called.
        It will create a new instance with sorted ``bonds`` and all the other
        attributes will be copied (through deepcopy).

        Returns
        -------
        undoubled_model : ExchangeModel
            Exchange model after reoving double bonds.
        """
        unbounded_model = deepcopy(self)
        for atom1 in self.bonds:
            for atom2 in self.bonds[atom1]:
                if atom1 in unbounded_model.bonds and\
                   atom2 in unbounded_model.bonds[atom1] and\
                   (0, 0, 0) in unbounded_model.bonds[atom1][atom2] and\
                   atom2 in unbounded_model.bonds and\
                   atom1 in unbounded_model.bonds[atom2] and\
                   (0, 0, 0) in unbounded_model.bonds[atom2][atom1]:
                    unbounded_model.bonds[atom1][atom2][(0, 0, 0)] = (unbounded_model.bonds[atom1][atom2][(0, 0, 0)] +
                                                                      unbounded_model.bonds[atom2][atom1][(0, 0, 0)]) * 0.5
                    unbounded_model.remove_bond(atom2, atom1, (0, 0, 0))
        return unbounded_model

    def filter(self,
               max_distance=None,
               min_distance=None,
               template=None,
               R_vector=None):
        """
        Filter the exchange entries based on the given conditions.

        The result will be defined by logical conjugate of the conditions.
        Saying so the filtering will be performed for each given condition
        one by one.
        Note: this method is not modifying the instance at which it is called.
        It will create a new instance with sorted :py:attr:`bonds` and all the other
        attributes will be copied (through :py:func:`deepcopy`).

        Parameters
        ----------

        max_distance : float or int, optional
            Distance for sorting, the condition is <<less or equal>>.

        min_distance : float or int, optional
            Distance for sorting, the condition is <<more or equal>>.

        template : list
            List of pairs for keeping. Specifying atom1, atom2, R.
            [(atom1, atom2, R), ...]

        R_vector : tuple of ints or list of tuples of ints
            Tuple of 3 integers or list of tuples, specifying the R vectors,
            which will be kept after filtering.

        Returns
        -------

        filtered_model : ExchangeModel
            Exchange model after filtering.
        """
        filtered_model = deepcopy(self)

        if template is not None:
            filtered_model.bonds = {}
            for atom1, atom2, R in template:
                filtered_model.add_bond(self.bonds[atom1][atom2][R],
                                        atom1, atom2, R)
        bonds_for_removal = set()
        if R_vector is not None:
            if type(R_vector) == tuple:
                R_vector = {R_vector}
            elif type(R_vector) == list:
                R_vector = set(R_vector)
        if min_distance is not None or\
           max_distance is not None or\
           R_vector is not None:
            for atom1 in filtered_model.bonds:
                for atom2 in filtered_model.bonds[atom1]:
                    for R in filtered_model.bonds[atom1][atom2]:

                        dis = filtered_model.bonds[atom1][atom2][R].dis

                        if max_distance is not None and dis > max_distance:
                            bonds_for_removal.add((atom1, atom2, R))

                        if min_distance is not None and dis < min_distance:
                            bonds_for_removal.add((atom1, atom2, R))

                        if R_vector is not None and not R in R_vector:
                            bonds_for_removal.add((atom1, atom2, R))

        for atom1, atom2, R in bonds_for_removal:
            filtered_model.remove_bond(atom1, atom2, R)
        return filtered_model


class ExchangeModelTB2J(ExchangeModel):
    """
    Class containing the exchange model extracted from TB2J output file.

    Parameters
    ----------

    filename : str
        Path to the TB2J output file.
    """

    # Flags for reading from TB2J file
    _major_sep = '=' * 90
    _minor_sep = '-' * 88
    _garbage = str.maketrans({'(': None,
                              ')': None,
                             '[': None,
                              ']': None,
                              ',': None,
                              '\'': None})
    _cell_flag = 'Cell (Angstrom):'
    _atoms_flag = 'Atoms:'
    _atom_end_flag = 'Total'
    _exchange_flag = 'Exchange:'
    _iso_flag = 'J_iso:'
    _aniso_flag = 'J_ani:'
    _dmi_flag = 'DMI:'

    # Store info about all atoms, not only magnetic ones
    # {mark: (x, y, z), ...}
    _atoms = {}

    def __init__(self, filename) -> None:
        super().__init__()
        file = open(filename, 'r')
        line = True
        self.file_order = []

        # Read everything before exchange
        while line:
            line = file.readline()

            # Read cell
            if line and self._cell_flag in line:
                self.cell = np.array([
                    list(map(float, file.readline().split())),
                    list(map(float, file.readline().split())),
                    list(map(float, file.readline().split()))
                ])

            # Read atoms
            if line and self._atoms_flag in line:
                line = file.readline()
                line = file.readline()
                line = file.readline().split()
                while line and self._atom_end_flag not in line:
                    self._atoms[line[0]] = tuple(map(float, line[1:4]))
                    line = file.readline().split()

            # Check if the exchange section is reached
            if line and self._exchange_flag in line:
                break

        # Read exchange
        while line:
            while line and self._minor_sep not in line:
                line = file.readline()
            line = file.readline().translate(self._garbage).split()
            atom1 = line[0]
            atom2 = line[1]
            R = tuple(map(int, line[2:5]))
            self.file_order.append((atom1, atom2, R))
            distance = float(line[-1])
            iso = None,
            aniso = None
            dmi = None
            while line and self._minor_sep not in line:
                line = file.readline()

                # Read isotropic exchange
                if line and self._iso_flag in line:
                    iso = float(line.split()[-1])

                # Read anisotropic exchange
                if line and self._aniso_flag in line:
                    aniso = np.array([
                        list(map(float, file.readline().translate(
                            self._garbage).split())),
                        list(map(float, file.readline().translate(
                            self._garbage).split())),
                        list(map(float, file.readline().translate(
                            self._garbage).split()))
                    ])

                # Read DMI
                if line and self._dmi_flag in line:
                    dmi = tuple(map(float, line.translate(
                        self._garbage).split()[-3:]))

            # Adding info from the exchange block to the ExchangeModel structure
            if atom1 not in self.magnetic_atoms:
                self.add_atom(atom1, *self._atoms[atom1])
            if atom2 not in self.magnetic_atoms:
                self.add_atom(atom2, *self._atoms[atom2])
            bond = Bond(iso=iso, aniso=aniso, dmi=dmi, distance=distance)
            self.add_bond(bond, atom1, atom2, R)

    # TODO Think about the class type problem
    def filter(self,
               max_distance=None,
               min_distance=None,
               template=None,
               R_vector=None):
        """
        Filter the exchange entries based on the given conditions.

        Call :py:meth:`ExchangeModel.filter` method from parent class and
        update :py:attr:`file_order`.
        """
        filtered_model = deepcopy(self)
        result = super().filter(max_distance, min_distance, template, R_vector)
        filtered_model.file_order = []
        filtered_model.bonds = result.bonds
        filtered_model.cell = result.cell
        filtered_model.magnetic_atoms = result.magnetic_atoms
        bond_list = set(filtered_model.get_bond_list())
        for key in self.file_order:
            if key in bond_list:
                filtered_model.file_order.append(key)
        return filtered_model



class ExchangeTemplate:
    """
    Store a template for sorting the exchange from TB2J file.

    In addition stores the technical details for plotting, 
    orbital decompozition, etc.

    Parameters
    ----------
    filename : str
        Path to the template file.

    Attributes
    ----------
    names : dict
        Dictionary of neighbours from the template file.

        .. code-block:: python

            {name : [(atom1, atom2, R), ...], ...}

    latex_names : dict
        The dictionary of Letex version of names from `names`.

        .. code-block:: python

            {name : latex_name, ...}
    """

    # Constants
    _major_sep = '=' * 20
    _minor_sep = '-' * 20
    _neighbors_flag = 'Neighbors template:'

    def __init__(self, filename) -> None:

        self.names = {}
        self.latex_names = {}
        with open(filename) as file:
            line = True
            while line:
                line = file.readline()

                # Read the neighbors names
                if line and self._neighbors_flag in line:
                    while line and self._major_sep not in line:
                        if line and self._minor_sep in line:
                            line = file.readline()
                            name = line.split()[0]
                            try:
                                latex_name = line.split()[1]
                            except IndexError:
                                latex_name = name
                            self.names[name] = []
                            self.latex_names[name] = latex_name
                            line = file.readline()
                            while line and\
                                    self._minor_sep not in line and\
                                    self._major_sep not in line:
                                atom1 = line.split()[0]
                                atom2 = line.split()[1]

                                R = tuple(map(int, line.split()[2:]))
                                self.names[name].append((atom1, atom2, R))
                                line = file.readline()
                        if line and self._major_sep in line:
                            break
                        if line and self._minor_sep not in line:
                            line = file.readline()

    def get_list(self):
        """
        Translate bonds present in the template into the list.

        Returns
        -------
        plained_template : list
            List with the bond specifications:

            .. code-block:: python

                [(atom_1, atom_2, R), ...]
        """
        plained_template = []
        for name in self.names:
            for atom1, atom2, R in self.names[name]:
                plained_template.append((atom1, atom2, R))
        return plained_template


if __name__ == "__main__":
    tmp = ExchangeTemplate(
        # "/Users/rybakov.ad/Projects/rad-tools/utest/tb2j_tools/resourses/template.txt")
        "/Users/rybakov.ad/Projects/rad-tools/debug/nii2/template.txt")
    print('here')



"""
Collection of small routines and constants, 
which may be used across the whole package.
"""

import numpy as np
from math import asin, sqrt, pi

# Teminal colors
BLACK = "\u001b[30m"
"""
ANSI escape code for black color of the text.
"""
RED = "\u001b[31m"
"""
ANSI escape code for red color of the text.
"""
GREEN = "\u001b[32m"
"""
ANSI escape code for green color of the text.
"""
YELLOW = "\u001b[33m"
"""
ANSI escape code for yellow color of the text.
"""
BLUE = "\u001b[34m"
"""
ANSI escape code for blue color of the text.
"""
MAGENTA = "\u001b[35m"
"""
ANSI escape code for magenta color of the text.
"""
CYAN = "\u001b[36m"
"""
ANSI escape code for cyan color of the text.
"""
WHITE = "\u001b[37m"
"""
ANSI escape code for white color of the text.
"""
RESET = "\u001b[0m"
"""
ANSI escape code for resetting color to default.
"""
WARNING = YELLOW
"""
ANSI escape code for warnings.
"""
OK = GREEN
"""
ANSI escape code for ok messages.
"""
ERROR = RED
"""
ANSI escape code for errors.
"""


def get_256_colours(n):
    """
    ANSI escape codes for terminal color with 256-colours support
    (see: https://en.wikipedia.org/wiki/ANSI_escape_code).

    Parameters
    ----------
    n : int
        Integer from 0 to 255 to be mapped to colours.

    Returns
    -------
    str
        string with the ANSI escape code.
    """

    if type(n) != int or not 0 <= n <= 255:
        raise ValueError(f'Integer n have to be in range 0 <= n <= 255. '
                         f'You provided n = {n}, type<n> = {type(n)}')
    return f'\033[38:5:{n}m'


def exchange_to_matrix(iso=None, aniso=None, dmi=None):
    """
    Combine isotropic, anisotropic and dmi exchange into exchange matrix.

    Parameters
    ----------
    iso : float
        Value of isotropic exchange parameter in meV.
    aniso : 3 x 3 array_like
        Matrix of symmetric anisotropic exchange in meV.
    dmi : 1 x 3 array_like
        Dzyaroshinsky-Moria interaction vector (Dx, Dy, Dz) in meV.

    Returns
    -------
    matrix : 3 x 3 np.ndarray of floats
        Exchange matrix in meV.
    """
    matrix = np.zeros((3, 3), dtype=float)
    if aniso is not None:
        matrix += np.array(aniso, dtype=float)
    if iso is not None:
        matrix += iso * np.identity(3, dtype=float)
    if dmi is not None:
        matrix += np.array([[0, dmi[2], -dmi[1]],
                            [-dmi[2], 0, dmi[0]],
                            [dmi[1], -dmi[0], 0]],
                           dtype=float)
    return matrix


def exchange_from_matrix(matrix):
    """
    Decompose matrix into isotropic, anisotropic and dmi exchange.

    Parameters
    ----------
    matrix : 3 x 3 array_like
        Exchange matrix in meV.

    Returns
    -------
    iso : float
        Value of isotropic exchange parameter in meV.
    aniso : 3 x 3 np.ndarray of floats
        Matrix of symmetric anisotropic exchange in meV.
    dmi : 1 x 3 np.ndarray of floats
        Dzyaroshinsky-Moria interaction vector (Dx, Dy, Dz) in meV.
    """
    matrix = np.array(matrix, dtype=float)
    symm = (matrix + matrix.T) / 2
    assym = (matrix - matrix.T) / 2
    dmi = np.array([assym[1][2], assym[2][0], assym[0][1]],
                   dtype=float)
    iso = np.trace(symm) / 3
    aniso = symm - iso * np.identity(3, dtype=float)
    return iso, aniso, dmi


def atom_mark_to_latex(mark):
    """
    Latexifier for atom marks. 

    Cr12 -> Cr\ :sub:`12`\.

    Parameters
    ----------
    mark : str
        Mark of atom.

    Returns
    -------
    new_mark : str
        Latex version of the mark.
    """
    numbers = '0123456789'
    new_mark = '$'
    insert_underline = False
    for symbol in mark:
        if symbol in numbers and not insert_underline:
            insert_underline = True
            new_mark += '_{'
        new_mark += symbol
    new_mark += '}$'
    return new_mark


def rot_angle(x, y, dummy=False):
    """
    Rotational ange from 2D vector.

    Mathematically positive => counterclockwise.
    From [0 to 360)

    Parameters
    ----------
    x : float or int
        x coordinate of a vector.
    y : float or int
        y coordinate of a vector.
    """
    # rot_cos = x / (x ** 2 + y ** 2) ** 0.5
    # rot_angle = m.acos(rot_cos) / m.pi * 180
    try:
        sin = abs(y) / sqrt(x**2 + y**2)
    except ZeroDivisionError:
        raise ValueError('Angle is ill defined (x = y = 0).')
    if x > 0:
        if y > 0:
            return asin(sin) / pi * 180
        elif y == 0:
            return 0
        elif y < 0:
            if not dummy:
                return -asin(sin) / pi * 180
            return 360 - asin(sin) / pi * 180
    elif x == 0:
        if y > 0:
            return 90
        elif y == 0:
            raise ValueError('Angle is ill defined (x = y = 0).')
        elif y < 0:
            if not dummy:
                return 90
            return 270
    elif x < 0:
        if y > 0:
            if not dummy:
                return -asin(sin) / pi * 180
            return 180 - asin(sin) / pi * 180
        elif y == 0:
            if not dummy:
                return 0
            return 180
        elif y < 0:
            if not dummy:
                return asin(sin) / pi * 180
            return 180 + asin(sin) / pi * 180



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os.path import join, abspath
from os import makedirs
from math import sqrt

import numpy as np

from rad_tools.exchange.model import ExchangeModelTB2J
from rad_tools.exchange.template import ExchangeTemplate
from rad_tools.routines import OK, RESET


def main(filename, out_dir, out_name, template):

    model = ExchangeModelTB2J(filename)
    template = ExchangeTemplate(template)
    with open(join(out_dir, out_name), "w") as out_file:
        for name in template.names:
            J_iso = 0
            J_aniso = np.zeros((3, 3), dtype=float)
            DMI = np.zeros(3, dtype=float)
            abs_DMI = 0
            for bond in template.names[name]:
                atom1 = bond[0]
                atom2 = bond[1]
                R = bond[2]
                J_iso += model.bonds[atom1][atom2][R].iso
                J_aniso += model.bonds[atom1][atom2][R].aniso
                DMI += model.bonds[atom1][atom2][R].dmi
                abs_DMI += sqrt(model.bonds[atom1][atom2][R].dmi[0]**2 +
                                model.bonds[atom1][atom2][R].dmi[1]**2 +
                                model.bonds[atom1][atom2][R].dmi[2]**2)
            J_iso /= len(template.names[name])
            J_aniso /= len(template.names[name])
            DMI /= len(template.names[name])
            abs_DMI /= len(template.names[name])
            out_file.write(f"{name}\n" +
                           f"    Isotropic: {round(J_iso, 4)}\n" +
                           f"    Anisotropic:\n" +
                           f"        {round(J_aniso[0][0], 4)} " +
                           f"    {round(J_aniso[0][1], 4)} " +
                           f"    {round(J_aniso[0][2], 4)}\n" +
                           f"        {round(J_aniso[1][0], 4)} " +
                           f"    {round(J_aniso[1][1], 4)} " +
                           f"    {round(J_aniso[1][2], 4)}\n" +
                           f"        {round(J_aniso[2][0], 4)} " +
                           f"    {round(J_aniso[2][1], 4)} " +
                           f"    {round(J_aniso[2][2], 4)}\n" +
                           f"    DMI: " +
                           f"    {round(DMI[0], 4)} " +
                           f"    {round(DMI[1], 4)} " +
                           f"    {round(DMI[2], 4)}\n" +
                           f"    | DMI |: {round(abs_DMI, 4)}\n" +
                           f"    | DMI/J | {round(abs(abs_DMI/J_iso), 4)}\n\n")

    print(f"{OK}Extracted exchange info is in " +
          f"{abspath(join(out_dir, out_name))}{RESET}")


if __name__ == '__main__':
    parser = ArgumentParser(
        description="Script for extracting of template-based model from TB2J results.",
        epilog="""
            For the full description of arguments see the docs:
            https: // rad-tools.adrybakov.com/en/stable/user-guide/tb2j-refractor.html
            """)

    parser.add_argument("-f", "--filename",
                        type=str,
                        required=True,
                        help="""
                        Relative or absulute path to the * exchange.out * file,
                        including the name and extention of the file itself.
                        """
                        )
    parser.add_argument("-t", "--template",
                        type=str,
                        required=True,
                        help="""
                        Relative or absolute path to the template file,
                        including the name and extention of the file.
                        """)
    parser.add_argument("-op", "--output-dir",
                        type=str,
                        default='.',
                        help="""
                        Relative or absolute path to the folder
                        for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default='exchange_refr.txt',
                        help="""
                        Seedname for the output files.
                        """
                        )

    args = parser.parse_args()

    try:
        makedirs(args.output_dir)
    except FileExistsError:
        pass

    main(filename=args.filename,
         out_dir=args.output_dir,
         out_name=args.output_name,
         template=args.template)



from random import randint

from rad_tools.routines import GREEN, RESET


class Action:

    exits = {
        "exit",
        "exit()",
        "quit",
        "quit()",
        "kill",
        "shutdown"}
    tb2j = {
        "TB2J",
        "tb2j"
    }
    confusions = [
        "You`re confused?\nI`m fucking confused, bro!",
        "Sorry, I did not understand you, could you try again?",
        "Jarvis does not support this type of input, try better.",
    ]

    def __init__(self, name) -> None:
        self.name = name
        self.decide_what_to_do()

    def decide_what_to_do(self):
        if self.name in self.exits:
            return 0
        elif self.name in self.tb2j:
            print(f"{GREEN}Start to set up TB2J calculation{RESET}")
            tb2j_action = TB2JAction("Begin")
        elif self.name == "Initiate":
            pass
        else:
            self.confused()
        self.name = self.listen_to_user()

    def listen_to_user(self):
        message = input()
        self.name = message
        self.decide_what_to_do()

    def confused(self):
        i = randint(0, len(self.confusions) - 1)
        print(i)
        print(f"{GREEN}{self.confusions[i]}{RESET}")


class TB2JAction(Action):

    def __init__(self, name) -> None:
        super().__init__(name)



#! /usr/local/bin/python3

from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np
import yaml


def plot(file, kp=None, constant=1, on="output"):
    data_file = open(file, 'r')
    data = yaml.load(data_file, Loader=yaml.FullLoader)
    data_file.close()
    X = []
    Y = []
    K_points = []
    miny = None
    maxy = None
    for i, point in enumerate(data['phonon']):
        for j, band in enumerate(data['phonon'][i]['band']):
            data['phonon'][i]['band'][j]['frequency'] *= constant
    for i, point in enumerate(data['phonon']):
        print(i)
        X.append(point['distance'])
        Y.append(list(map(lambda x: x['frequency'], point['band'])))
        if i == 0 or i == len(data['phonon']) - 1 or data['phonon'][i]['q-position'] == data['phonon'][i + 1]['q-position']:
            K_points.append(point['distance'])
        if not miny:
            miny = min(Y[-1])
        if not maxy:
            maxy = max(Y[-1])
        miny = min(min(Y[-1]), miny)
        maxy = max(max(Y[-1]), maxy)

    print(K_points)

    fig, ax = plt.subplots()
    X = np.array(X)
    Y = np.array(Y)
    ax.plot(X, Y, color='red')
    ax.set_ylabel('Phonon frequency (THz)', fontsize=15)
    ax.set_xlim(X[0], X[-1])
    if kp is None or len(kp) != len(K_points):
        if kp is None:
            kp = []
        print(f'\033[93m'
              f'{len(K_points)} k-points expected, only {len(kp)} k-points are provided. '
              f'Please, check the input.'
              f'\nProvided K-points: {kp}'
              f'\x1b[0m')
    if kp and len(kp) == len(K_points):
        for i, k in enumerate(kp):
            if k == 'G':
                kp[i] = '$\Gamma$'
        ax.xaxis.set_ticks(K_points, kp, fontsize=15)
        for k_point in K_points:
            ax.vlines(k_point, 0, 1, transform=ax.get_xaxis_transform(),
                      color='black', linewidths=0.5)
    plt.show()
    if constant != 1:
        output = yaml.dump(data)
        with open(f"{on}.yaml", "w") as file:
            file.write(output)


if __name__ == '__main__':
    parser = ArgumentParser(description='Fit Exchange parameters map')

    parser.add_argument("-file", type=str, default='band.yaml',
                        help="Path to the band.yaml file")
    parser.add_argument("-kp",  type=str, default=None,
                        nargs='*',
                        help="K-path, example: -kp G X R Y G")
    parser.add_argument('-constant', type=float, default=1,
                        help="Constant for conversion. "
                        "The values will be multiplied by it")
    parser.add_argument('-on', type=str, default="output",
                        help="output name file")
    args = parser.parse_args()
    if args.file:
        plot(args.file, kp=args.kp, constant=args.constant, on=args.on)



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os.path import join, abspath
from os import makedirs

import numpy as np

from rad_tools.exchange.model import ExchangeModelTB2J
from rad_tools.routines import OK, RESET, YELLOW


def provide_template(out_dir=".", out_name="template.txt",
                     tb2j_filename=None,
                     min_distance=None,
                     max_distance=None,
                     R_vector=None):
    template = ('=' * 20 + "\n" +
                "Neighbors template:\n" +
                "i j R_a R_b R_c\n" +
                '-' * 20 + "\n" +
                "J1 $J_1$\n" +
                "atom1 atom2 0 0 0\n" +
                "atom1 atom2 1 0 0\n" +
                "atom1 atom1 -1 0 2\n" +
                '-' * 20 + "\n" +
                "J2\n" +
                "atom2 atom1 9 5 -3\n" +
                "atom1 atom2 1 4 0\n" +
                "atom2 atom2 1 0 2\n" +
                '=' * 20 + "\n")

    with open(join(out_dir, out_name), "w") as file:
        if tb2j_filename is None:
            file.write(template)
        else:
            model = ExchangeModelTB2J(tb2j_filename)
            model = model.filter(min_distance=min_distance,
                                 max_distance=max_distance,
                                 R_vector=R_vector)
            file .write('=' * 20 + "\n" +
                        "Neighbors template:\n" +
                        "i j R_a R_b R_c\n" +
                        '-' * 20 + "\n")
            for atom1, atom2, R in model.file_order:
                file.write(f"{atom1} {atom2} {R[0]} {R[1]} {R[2]}\n")
            file .write('=' * 20 + "\n")
    print(f"{OK}Template draft is in " +
          f"{abspath(join(out_dir, out_name))}{RESET}")
    print(f"{YELLOW}Do not forget to correct the template draft " +
          f"to your needs!{RESET}")


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Script for the creation of template`s template",
        epilog="""
            For the full description of arguments see the docs:
            https://rad-tools.adrybakov.com/en/stable/user-guide/rad-make-template.html
            """)

    parser.add_argument("-op", "--output-dir",
                        type=str,
                        default='.',
                        help="""
                        Relative or absolute path to the folder for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default='template.txt',
                        help="""
                        Template file name, default "template.txt"
                        """
                        )
    parser.add_argument("-f", "--filename",
                        type=str,
                        default=None,
                        help="""
                        Relative or absulute path to the *exchange.out* file,
                        including the name and extention of the file itself.
                        """
                        )
    parser.add_argument("-R", "--R-vector",
                        type=int,
                        nargs="*",
                        default=None,
                        help="""
                        R vectors for filtering the model.
                        """
                        )
    parser.add_argument("-maxd", "--max-distance",
                        type=float,
                        default=None,
                        help="""
                        (<=) Maximum distance.
                        """
                        )
    parser.add_argument("-mind", "--min-distance",
                        type=float,
                        default=None,
                        help="""
                        (>=) Minimum distance.
                        """
                        )
    parser.add_argument("-d", "--distance",
                        type=float,
                        default=None,
                        help="""
                        (=) Exact distance.
                        """
                        )

    args = parser.parse_args()

    try:
        makedirs(args.output_dir)
    except FileExistsError:
        pass

    if args.distance is not None:
        args.min_distance = args.max_distance = args.distances

    if args.R_vector is not None:
        args.R_vector = np.array(args.R_vector[:len(args.R_vector) // 3 * 3],
                                 dtype=int).reshape((len(args.R_vector)//3, 3))
        args.R_vector = list(map(tuple, args.R_vector.tolist()))

    provide_template(out_dir=args.output_dir,
                     out_name=args.output_name,
                     tb2j_filename=args.filename,
                     min_distance=args.min_distance,
                     max_distance=args.max_distance,
                     R_vector=args.R_vector)



from setuptools import setup, find_packages
from rad_tools import __version__

__version__ = __version__

with open('README.rst', 'r') as file:
    long_description = ''
    for line in file:
        long_description += line

setup(
    name='rad-tools',
    version=__version__,
    description='Collection of scripts from my PhD',
    long_description=long_description,
    author='Andrey Rybakov',
    author_email='rybakov.ad@icloud.com',
    license='MIT license',
    url='https://github.com/adrybakov/rad-tools',
    download_url='https://github.com/adrybakov/rad-tools.git',
    packages=find_packages(),
    scripts=[
        'scripts/tb2j-plotter.py',
        'scripts/phonopy-plotter.py',
        'scripts/tb2j-refractor.py',
        'scripts/rad-make-template.py',
        'scripts/rad-jarvis.py',
        'scripts/rad-dos-plotter.py'
    ],
    install_requires=[
        'numpy', 'matplotlib'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.6',
)



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os import makedirs
from os.path import basename, abspath, join

from rad_tools.routines import OK, RESET

import matplotlib.pyplot as plt
import numpy as np

LABELS = {
    "s": ["s (up)", "s (down)"],
    "p": ["$p_z$ (up)", "$p_z$ (down)",
          "$p_y$ (up)", "$p_y$ (down)",
          "$p_x$ (up)", "$p_x$ (down)"],
    "d": ["$d_{z^2}$ (up)", "$d_{z^2}$",
          "$d_{zx}$ (up)", "$d_{zx}$ (down)",
          "$d_{zy}$ (up)", "$d_{zy}$ (down)",
          "$d_{x^2 - y^2}$ (up)", "$d_{x^2 - y^2}$ (down)",
          "$d_{xy}$ (up)", "$d_{xy}$ (down)"]
}

COLORS = {
    "s": ["blue", "red"],
    "p": ["#FF5C5C", "#6B0000",
          "#37FF5C", "#00A61E",
          "#2591FF", "#0061C4"],
    "d": ["#FF297A", "#6B0029",
          "#FFF821", "#898500",
          "#19FF19", "#00A600",
          "#10F9FF", "#00BFC4",
          "#5A08FF", "#4B00E1"]
}


def read_dos(filename, efermi=0):
    dos = np.loadtxt(filename, skiprows=1).T
    with open(filename, "r") as file:
        header = file.readline()
    dos[0] -= efermi
    if "up" in header and "dw" in header:
        for i in range(0, (len(dos) - 1) // 2):
            dos[i * 2 + 2] *= -1
    return dos


def decompose_filename(filename):
    atom_type = filename.split("(")[1].split(")")[0]
    atom_number = filename.split("#")[1].split("(")[0]
    orbital_type = filename.split("(")[2].split(")")[0]
    orbital_number = filename.split("#")[2].split("(")[0]
    return atom_type, orbital_type, atom_number, orbital_number


def manager(filename, out_dir, out_name, window=None):
    atom_type, orbital_type, atom_number, orbital_number = decompose_filename(
        basename(filename))
    dos = read_dos(filename)

    fig, ax = plt.subplots(figsize=(12, 4))

    ax.set_xlabel("Energy, eV")
    ax.set_ylabel("DOS, states/eV")
    ax.set_title(f"Atom: {atom_type}, orbitals: {orbital_type}")
    ax.vlines(0, 0, 1, transform=ax.get_xaxis_transform(),
              color='grey', linewidths=0.5,
              linestyles='dashed')

    ax.fill_between(dos[0], dos[2], dos[1],
                    color='grey', alpha=0.5, lw=0)
    # plot up
    for i in range(0, (len(dos) - 3) // 2):
        ax.plot(dos[0], dos[3 + 2 * i],
                '-',
                c=COLORS[orbital_type][2 * i],
                label=LABELS[orbital_type][2 * i],
                linewidth=0.7)
    # plot down
    for i in range(0, (len(dos) - 3) // 2):
        ax.plot(dos[0], dos[4 + 2 * i],
                '-',
                c=COLORS[orbital_type][2 * i + 1],
                label=LABELS[orbital_type][2 * i + 1],
                linewidth=0.7)

    ax.legend(loc='best', ncol=2)

    if window is not None:
        ax.set_xlim(*tuple(window))

    png_path = join(
        out_dir, f"{out_name}.#{atom_number}({atom_type})_wfc#{orbital_number}({orbital_type}).png")
    plt.savefig(png_path, dpi=400)
    print(f'{OK}Plot is in {abspath(png_path)}{RESET}')


if __name__ == "__main__":
    parser = ArgumentParser(description="Script for visualisation of DOS",
                            epilog="""
               For the full description of arguments see the docs: 
               https://rad-tools.adrybakov.com/en/stable/user-guide/rad-dos-plotter.html
               """)
    parser.add_argument("-f", "--filename",
                        type=str,
                        required=True,
                        help="""
                        Relative or absulute path to the dos file,
                        including the name and extention of the file itself.
                        """
                        )
    parser.add_argument("-op", "--output-dir",
                        type=str,
                        default='.',
                        help="""
                        Relative or absolute path to the folder
                        for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default='pdos',
                        help="""
                        Seedname for the output files.
                        """
                        )
    parser.add_argument("-w", "--window",
                        type=float,
                        nargs=2,
                        default=None,
                        help="""
                        Energy window.
                        """)

    args = parser.parse_args()

    try:
        makedirs(args.output_dir)
    except FileExistsError:
        pass

        manager(filename=args.filename,
                out_dir=args.output_dir,
                out_name=args.output_name,
                window=args.window)



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os.path import join, abspath
from os import makedirs
from math import sqrt

from matplotlib import pyplot as plt
import matplotlib as mpl
import numpy as np

from rad_tools.exchange.model import ExchangeModelTB2J
from rad_tools.routines import atom_mark_to_latex, rot_angle, OK, RESET


def plot_2d(filename, out_dir='.',
            out_name='exchange',
            wtp='iso',
            draw_cells=False,
            min_distance=None,
            max_distance=None,
            template=None,
            R_vector=None,
            double_bonds=False,
            scale_atoms=1,
            scale_data=1,
            atoms=None,
            title=None):

    mode_name = "2d"
    messages = {'iso': "isotropic exchange",
                "distance": "distances"}

    model = ExchangeModelTB2J(filename)
    model = model.filter(min_distance=min_distance,
                         max_distance=max_distance,
                         R_vector=R_vector,
                         template=template)
    if atoms is None:
        atoms = model.magnetic_atoms

    dummy = True
    ha = 'right'
    if not double_bonds:
        model = model.remove_double_bonds()
        dummy = False
        ha = 'center'
    X, Y, Z = model.get_space_dimensions()
    if X == 0 and Y == 0:
        X = Y = 1
    fontsize = 10 * 1.1 * Y / 5
    plt.rcParams.update({'font.size': fontsize})
    mpl.rcParams.update({'axes.linewidth': 1.1 * Y / 5})
    mpl.rcParams.update({'xtick.major.width': 1.1 * Y / 5})
    mpl.rcParams.update({'ytick.major.width': 1.1 * Y / 5})
    mpl.rcParams.update({'xtick.major.size': 5 * 1.1 * Y / 5})
    mpl.rcParams.update({'ytick.major.size': 5 * 1.1 * Y / 5})
    mpl.rcParams.update({'xtick.major.pad': 5 * 1.1 * Y / 5})
    mpl.rcParams.update({'ytick.major.pad': 5 * 1.1 * Y / 5})
    mpl.rcParams.update({'axes.labelpad': 1.1 * Y / 5})

    fig = plt.figure(figsize=(1.1 * X, 1.1 * Y))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    ax.set_xlabel('x, Angstroms')
    ax.set_ylabel('y, Angstroms')
    label_ax = fig.add_axes([0.85, 0.1, 0.1, 0.8])
    label_ax.set_xlim(0, 1)
    label_ax.set_ylim(0, 1)
    label_ax.axis('off')

    for atom1 in model.bonds:
        for atom2 in model.bonds[atom1]:
            for R in model.bonds[atom1][atom2]:
                bond = model.bonds[atom1][atom2][R]
                x1, y1, z1, x2, y2, z2 = model.get_atom_coordinates(atom1,
                                                                    atom2,
                                                                    R)
                xm = (x1 + x2) / 2
                ym = (y1 + y2) / 2
                zm = (z1 + z2) / 2

                ax.scatter(x1, y1, s=100 * fig.dpi/72., c='white')
                ax.scatter(x2, y2, s=100 * fig.dpi/72., c='white')

                ax.text(x1, y1, atom_mark_to_latex(atom1),
                        va='center', ha='center',
                        fontsize=1.5 * fontsize * scale_atoms)
                ax.text(x2, y2, atom_mark_to_latex(atom2),
                        va='center', ha='center',
                        fontsize=1.5 * fontsize * scale_atoms)
                if wtp == 'iso':
                    ax.text(xm, ym, str(round(bond.iso, 4)),
                            va='bottom', ha=ha,
                            rotation_mode='anchor',
                            rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                            fontsize=fontsize * scale_data)
                elif wtp == 'distance':
                    ax.text(xm, ym, str(round(bond.dis, 4)),
                            va='bottom', ha=ha,
                            rotation_mode='anchor',
                            rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                            fontsize=fontsize * scale_data)

    if draw_cells:
        xlims = ax.get_xlim()
        ylims = ax.get_ylim()
        cells = model.get_cells()
        a_x, a_y, a_z = tuple(model.cell[0])
        b_x, b_y, b_z = tuple(model.cell[1])
        Rx_min = 0
        Rx_max = 0
        Ry_min = 0
        Ry_max = 0
        Rz_min = 0
        Rz_max = 0
        for Rx, Ry, Rz in cells:
            Rx_min = min(Rx, Rx_min)
            Rx_max = max(Rx, Rx_max)
            Ry_min = min(Ry, Ry_min)
            Ry_max = max(Ry, Ry_max)
            Rz_min = min(Rz, Rz_min)
            Rz_max = max(Rz, Rz_max)

        for i in range(Rx_min, Rx_max + 1):
            for j in range(Ry_min, Ry_max + 1):
                for k in range(Rz_min, Rz_max + 1):
                    X_shift = i * a_x + j * b_x
                    Y_shift = i * a_y + j * b_y
                    ax.plot(np.array([0, a_x, a_x + b_x, b_x, 0]) + X_shift,
                            np.array([0, a_y, a_y + b_y, b_y, 0]) + Y_shift,
                            linewidth=1, color="#BCBF5A")
        ax.set_xlim(*xlims)
        ax.set_ylim(*ylims)

    if title is not None:
        ax.set_title(title, fontsize=1.5 * fontsize)

    png_path = join(out_dir, f'{out_name}.{wtp}.png')
    plt.savefig(png_path, dpi=400)
    print(f'{OK}{mode_name} plot with {wtp} is in {abspath(png_path)}{RESET}')


def get_molecule_center(atoms, exclude_atoms):
    filtered_atoms = {}
    for atom in atoms:
        keep = True
        for exclude_atom in exclude_atoms:
            if exclude_atom in atom:
                keep = False
                break
        if keep:
            filtered_atoms[atom] = atoms[atom]
    atoms = filtered_atoms
    x_min, y_min, z_min = None, None, None
    x_max, y_max, z_max = None, None, None
    for atom in atoms:
        x, y, z = atoms[atom]

        if x_min is None:
            x_min = x
        else:
            x_min = min(x, x_min)
        if y_min is None:
            y_min = y
        else:
            y_min = min(y, y_min)
        if z_min is None:
            z_min = z
        else:
            z_min = min(z, z_min)

        if x_max is None:
            x_max = x
        else:
            x_max = max(x, x_max)
        if y_max is None:
            y_max = y
        else:
            y_max = max(y, y_max)
        if z_max is None:
            z_max = z
        else:
            z_max = max(z, z_max)
            return (x_min + x_max) / 2, (y_min + y_max) / 2, (z_min + z_max) / 2


def get_distance(x1, y1, z1, x2, y2, z2):
    return sqrt((x2 - x1)**2 + (y2-y1)**2 + (z2-z1)**2)


def plot_molecule(filename, out_dir='.',
                  out_name='exchange',
                  wtp='iso',
                  draw_cells=False,
                  min_distance=None,
                  max_distance=None,
                  template=None,
                  R_vector=None,
                  double_bonds=False,
                  scale_atoms=1,
                  scale_data=1,
                  atoms=None,
                  title=None):

    mpl.rcParams.update(mpl.rcParamsDefault)
    mode_name = "molecule"
    messages = {'iso': "isotropic exchange",
                "distance": "distances"}

    model = ExchangeModelTB2J(filename)
    model = model.filter(min_distance=min_distance,
                         max_distance=max_distance,
                         R_vector=R_vector,
                         template=template)

    if atoms is None:
        atoms = model.magnetic_atoms

    fig, ax = plt.subplots()

    ax.set_xlabel('Distance to the molecule center, Angstroms')
    if wtp == 'iso':
        ax.set_ylabel('Isotropic exchange parameter, meV')
    elif wtp == 'distance':
        ax.set_ylabel('Bond length, Angstroms')

    data = [[], []]
    x, y, z = get_molecule_center(model._atoms, atoms)
    print(f"{OK}Molecule center is detected at "
          f"({round(x, 4)}, {round(y, 4)}, {round(z, 4)}){RESET}")

    for atom1 in model.bonds:
        for atom2 in model.bonds[atom1]:
            for R in model.bonds[atom1][atom2]:
                xa, ya, za = model.get_bond_coordinate(atom1, atom2, R)
                d = get_distance(x, y, z, xa, ya, za)
                bond = model.bonds[atom1][atom2][R]
                data[0].append(d)
                if wtp == 'iso':
                    data[1].append(bond.iso)
                elif wtp == 'distance':
                    data[1].append(bond.dis)
    data = np.array(data).T.tolist()

    data.sort(key=lambda x: x[0])
    data = np.array(data).T.tolist()
    ax.plot(data[0], data[1], "o-", color="black")
    if title is not None:
        ax.set_title(title, fontsize=15)

    png_path = join(out_dir, f'{out_name}.{mode_name}.{wtp}.png')
    plt.savefig(png_path, dpi=400)
    print(f'{OK}{mode_name} plot with {wtp} is in {abspath(png_path)}{RESET}')


if __name__ == '__main__':
    parser = ArgumentParser(
        description="Script for visualisation of TB2J results",
        epilog="""
               For the full description of arguments see the docs: 
               https://rad-tools.adrybakov.com/en/stable/user-guide/tb2j-plotter.html
               """)

    plot_data_type = ['iso', 'distance']
    plot_mode = {"2d": plot_2d, "molecule": plot_molecule}

    parser.add_argument("-f", "--filename",
                        type=str,
                        required=True,
                        help="""
                        Relative or absulute path to the *exchange.out* file,
                        including the name and extention of the file itself.
                        """
                        )
    parser.add_argument("-m", "--mode",
                        type=str,
                        default="2d",
                        choices=["all", "2d", "molecule"],
                        help="""
                        Mode of plotting.
                        """
                        )
    parser.add_argument("-a", "--atoms",
                        type=str,
                        default=None,
                        nargs="*",
                        help="""
                        Atoms from the substrate.
                        """
                        )
    parser.add_argument("-op", "--output-dir",
                        type=str,
                        default='.',
                        help="""
                        Relative or absolute path to the folder for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default='exchange',
                        help="""
                        Seedname for the output files.
                        """
                        )
    parser.add_argument("-wtp", "--what-to-plot",
                        type=str,
                        choices=['all'] + plot_data_type,
                        default='all',
                        help="""
                        Type of data for display.
                        """
                        )
    parser.add_argument("-dc", "--draw-cells",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to draw the supercell`s shape.
                        """
                        )
    parser.add_argument("-R", "--R-vector",
                        type=int,
                        nargs="*",
                        default=None,
                        help="""
                        R vectors for filtering the model.
                        """
                        )
    parser.add_argument("-maxd", "--max-distance",
                        type=float,
                        default=None,
                        help="""
                        (<=) Maximum distance.
                        """
                        )
    parser.add_argument("-mind", "--min-distance",
                        type=float,
                        default=None,
                        help="""
                        (>=) Minimum distance.
                        """
                        )
    parser.add_argument("-d", "--distance",
                        type=float,
                        default=None,
                        help="""
                        (=) Exact distance.
                        """
                        )
    parser.add_argument("-tf", "--template",
                        type=str,
                        default=None,
                        help="""
                        Relative or absolute path to the template file, 
                        including the name and extention of the file.
                        """)
    parser.add_argument("-db", "--double-bonds",
                        default=False,
                        action="store_true",
                        help="""
                        Whenever to keep both bonds.
                        """)
    parser.add_argument("-sa", "--scale-atoms",
                        default=1,
                        type=float,
                        help="""
                        Scale for the size of atom marks.
                        """)
    parser.add_argument("-sd", "--scale-data",
                        default=1,
                        type=float,
                        help="""
                        Scale for the size of data text.
                        """)
    parser.add_argument("-t", "--title",
                        default=None,
                        type=str,
                        help="""
                        Title for the plots.
                        """)

    args = parser.parse_args()

    if args.distance is not None:
        args.min_distance = args.max_distance = args.distance

    try:
        makedirs(args.output_dir)
    except FileExistsError:
        pass

    if args.R_vector is not None:
        args.R_vector = np.array(args.R_vector[:len(args.R_vector) // 3 * 3],
                                 dtype=int).reshape((len(args.R_vector)//3, 3))
        args.R_vector = list(map(tuple, args.R_vector.tolist()))

    if args.what_to_plot != 'all':
        plot_data_type = [args.what_to_plot]
    if args.mode != 'all':
        plot_mode = [plot_mode[key] for key in [args.mode]]
    else:
        plot_mode = [plot_mode[key] for key in plot_mode]

    for mode in plot_mode:
        for data_type in plot_data_type:
            mode(filename=args.filename,
                 out_dir=args.output_dir,
                 out_name=args.output_name,
                 wtp=data_type,
                 draw_cells=args.draw_cells,
                 min_distance=args.min_distance,
                 max_distance=args.max_distance,
                 template=args.template,
                 R_vector=args.R_vector,
                 double_bonds=args.double_bonds,
                 scale_atoms=args.scale_atoms,
                 scale_data=args.scale_data,
                 atoms=args.atoms,
                 title=args.title)





from tracemalloc import start


class CalculationMap:
    """
    Mapper.

    Attributes
    ----------
    parameters : dict
        Dictionary of the parameters.

        .. code-block:: python

            {name : values, ...}
    """

    def __init__(self) -> None:
        self.parameters = {}

    def add_parameter(self,
                      name,
                      values=None,
                      range=None):
        """
        Add parameter to the class.

        Parameters
        ----------
        name : str 
            The main identificator of the parameter.
        values : array-like
            exact values of the parameter to be considered.
        range : tuple 
            Three numbers: start, stop, step. 
            If `values` are not specified the range will be used to 
            create them. If (`stop` - `start`) != n * `step`, where n - int
            then the last point will be < `stop`.
        """
        if values is None:
            if range is None:
                raise ValueError("Either values or range need to be specified")
            else:
                values = []
                value = range[0]
                while value <= range[1]:
                    values.append(value)
                    value += range[2]

        self.parameters[name] = values
